<?php
// Own namespace
namespace CoreFramework\Filesystem\Pointer;

// Import framework stuff
use CoreFramework\Bootstrap\FrameworkBootstrap;
use CoreFramework\FileSystem\BaseFileIo;
use CoreFramework\FileSystem\FileReadProtectedException;
use CoreFramework\FileSystem\FileWriteProtectedException;
use CoreFramework\Filesystem\PathWriteProtectedException;
use CoreFramework\Generic\NullPointerException;
use CoreFramework\Generic\UnsupportedOperationException;
use CoreFramework\Object\BaseFrameworkSystem;

/**
 * A class for reading files
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2017 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class FrameworkFileInputOutputPointer extends BaseFileIo implements InputOutputPointer {
	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	protected function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Create a file pointer based on the given file. The file will also
	 * be verified here.
	 *
	 * @param	$fileName	The file name we shall pass to fopen()
	 * @return	void
	 * @throws	FileIsEmptyException		If the given file name is NULL or empty
	 * @throws	FileReadProtectedException	If PHP cannot read an existing file
	 * @throws	FileWriteProtectedException	If PHP cannot write an existing file
	 * @throws	PathWriteProtectedException	If PHP cannot write to an existing path
	 * @throws	FileIoException				If fopen() returns not a file resource
	 */
	public static final function createFrameworkFileInputOutputPointer ($fileName) {
		// Some pre-sanity checks...
		if ((is_null($fileName)) || (empty($fileName))) {
			// No filename given
			throw new FileIsEmptyException(NULL, self::EXCEPTION_UNEXPECTED_EMPTY_STRING);
		} elseif (!FrameworkBootstrap::isReachableFilePath($fileName)) {
			// File exists but cannot be read
			throw new FileIoException($fileName, self::EXCEPTION_FILE_NOT_REACHABLE);
		} elseif ((!FrameworkBootstrap::isReadableFile($fileName)) && (file_exists($fileName))) {
			// File exists but cannot be read
			throw new FileReadProtectedException($fileName, self::EXCEPTION_FILE_CANNOT_BE_READ);
		} elseif ((file_exists($fileName)) && (!is_writable($fileName))) {
			// File exists but cannot be written
			throw new FileWriteProtectedException($fileName, self::EXCEPTION_FILE_CANNOT_BE_WRITTEN);
		} elseif (!is_writable(dirname($fileName))) {
			// Path is not writable
			throw new PathWriteProtectedException($fileName, self::EXCEPTION_PATH_CANNOT_BE_WRITTEN);
		}

		// Try to open a handler
		$filePointer = fopen($fileName, 'c+b');
		if ((is_null($filePointer)) || ($filePointer === false)) {
			// Something bad happend
			throw new FileIoException($fileName, self::EXCEPTION_FILE_POINTER_INVALID);
		} // END - if

		// Create new instance
		$pointerInstance = new FrameworkFileInputOutputPointer();

		// Set file pointer and file name
		$pointerInstance->setPointer($filePointer);
		$pointerInstance->setFileName($fileName);

		// Return the instance
		return $pointerInstance;
	}

	/**
	 * Validates file pointer and throws exceptions. This method does not return
	 * anything (not reliable) as this method checks the file pointer and on
	 * case of an error it throws an exception. If this method does not throw
	 * any exceptions, the file pointer seems to be fine.
	 *
	 * @return	void
	 * @throws	NullPointerException	If the file pointer instance
	 *									is not set by setPointer()
	 * @throws	InvalidResourceException	If there is being set
	 */
	private function validateFilePointer () {
		if (is_null($this->getPointer())) {
			// Pointer not initialized
			throw new NullPointerException($this, self::EXCEPTION_IS_NULL_POINTER);
		} elseif (!is_resource($this->getPointer())) {
			// Pointer is not a valid resource!
			throw new InvalidResourceException($this, self::EXCEPTION_INVALID_RESOURCE);
		}

		// All fine here
	}

	/**
	 * Read 1024 bytes data from a file pointer
	 *
	 * @return	mixed	The result of fread()
	 */
	public function readFromFile () {
		// Validate the pointer
		$this->validateFilePointer();

		// Read data from the file pointer and return it
		return $this->read(1024);
	}

	/**
	 * Write data to a file pointer
	 *
	 * @param	$dataStream		The data stream we shall write to the file
	 * @return	mixed			Number of writes bytes or false on error
	 */
	public function writeToFile ($dataStream) {
		// Validate the pointer
		$this->validateFilePointer();

		// Write data to the file pointer and return written bytes
		return fwrite($this->getPointer(), $dataStream, strlen($dataStream));
	}

	/**
	 * Writes at given position by seeking to it.
	 *
	 * @param	$seekPosition	Seek position in file
	 * @param	$data			Data to be written
	 * @return	mixed			Number of writes bytes or false on error
	 */
	public function writeAtPosition ($seekPosition, $data) {
		// First seek to it
		$this->seek($seekPosition);

		// Then write the data at that position
		return $this->writeToFile($data);
	}

	/**
	 * Rewinds to the beginning of the file
	 *
	 * @return	$status		Status of this operation
	 */
	public function rewind () {
		// Validate the pointer
		$this->validateFilePointer();

		// Rewind the pointer
		return rewind($this->getPointer());
	}

	/**
	 * Seeks to given position
	 *
	 * @param	$seekPosition	Seek position in file
	 * @param	$whence			"Seek mode" (see http://de.php.net/fseek)
	 * @return	$status			Status of this operation
	 */
	public function seek ($seekPosition, $whence = SEEK_SET) {
		// Validate the pointer
		$this->validateFilePointer();

		// Move the file pointer
		return fseek($this->getPointer(), $seekPosition, $whence);
	}

	/**
	 * Reads a line, maximum 4096 Bytes from current file pointer
	 *
	 * @return	$data	Read data from file
	 */
	public function readLine () {
		// Read whole line
		return $this->read();
	}

	/**
	 * Reads given amount of bytes from file.
	 *
	 * @param	$bytes	Amount of bytes to read
	 * @return	$data	Data read from file
	 */
	public function read ($bytes = NULL) {
		// Validate the pointer
		$this->validateFilePointer();

		// Is $bytes set?
		if (is_int($bytes)) {
			// Try to read given characters
			$data = fread($this->getPointer(), $bytes);
		} else {
			// Try to read whole line
			$data = fread($this->getPointer());
		}

		// Then return it
		return $data;
	}

	/**
	 * Analyzes entries in index file. This will count all found (and valid)
	 * entries, mark invalid as damaged and count gaps ("fragmentation"). If
	 * only gaps are found, the file is considered as "virgin" (no entries).
	 *
	 * @return	void
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function analyzeFile () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Advances to next "block" of bytes
	 *
	 * @return	void
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function next () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Checks wether the current entry is valid (not at the end of the file).
	 * This method will return true if an emptied (nulled) entry has been found.
	 *
	 * @return	$isValid	Whether the next entry is valid
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function valid () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Gets current seek position ("key").
	 *
	 * @return	$key	Current key in iteration
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function key () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * "Getter" for file size
	 *
	 * @return	$fileSize	Size of currently loaded file
	 */
	public function getFileSize () {
		// Check if the pointer is still valid
		$this->validateFilePointer();

		// Get file's data
		$fileData = fstat($this->getPointer());

		// Make sure the required array key is there
		assert(isset($fileData['size']));

		// Return size
		return $fileData['size'];
	}

}
